#include <u.h>
#include <libc.h>
#include <bio.h>

typedef struct Map Map;
typedef char *Array;
typedef struct Opcode Opcode;

static struct Map {
	int id;					/* value */
	char *str;				/* name */
};

static struct Opcode {
	int id;					/* operator id */
	int nargs;				/* bytes of args */
	char *str;				/* text for debug */
};

enum {
	MTU = 16,									/* maximum message length */
};

enum {
	Broadcast = 0x0f,
};

enum {
	CecFeatureAbort						= 0x00,
	CecSetAnaloguePreset				= 0x01,		/* v1.2,		removed in 1.3a */
	CecStartPresetTransfer				= 0x02,		/* v1.2,		removed in 1.3a */
	CecEndPresetTransfer				= 0x03,		/* v1.2,		removed in 1.3a */
	CecImageViewOn						= 0x04,
	CecTunerStepIncrement				= 0x05,
	CecTunerStepDecrement				= 0x06,
	CecTunerDeviceStatus				= 0x07,
	CecGiveTunerDeviceStatus			= 0x08,
	CecRecordOn							= 0x09,
	CecRecordStatus						= 0x0a,
	CecRecordOff						= 0x0b,
	CecSetPresetName					= 0x0c,		/* v1.2,		removed in 1.3a */
	CecTextViewOn						= 0x0d,
	CecRecordTvScreen					= 0x0f,
	CecStopPresetTransfer				= 0x10,		/* v1.2,		removed in 1.3a */
	CecSelectPreset						= 0x13,		/* v1.2,		removed in 1.3a */
	CecGivePresetListVersionNumber		= 0x15,		/* v1.2,		removed in 1.3a */
	CecActualPresetListVersionNumber	= 0x16,		/* v1.2,		removed in 1.3a */
	CecTimerRecordListTrigger			= 0x19,		/* v1.2,		removed in 1.3a */
	CecGiveDeckStatus					= 0x1a,
	CecDeckStatus						= 0x1b,
	CecSetCountry						= 0x31,		/* v1.2,		removed in 1.3a */
	CecSetLanguage						= 0x32,		/* v1.2,		renamed in 1.3a */
	CecSetMenuLanguage					= 0x32,		/* 1.3a */
	CecClearTimer						= 0x33,		/* v1.2,		renamed in 1.3a */
	CecClearAnalogueTimer				= 0x33,		/* 1.3a */
	CecSetTimer							= 0x34,		/* v1.2,		renamed in 1.3a */
	CecSetAnalogueTimer					= 0x34,		/* 1.3a */
	CecTimerStatus						= 0x35,
	CecStandby							= 0x36,
	CecStartTimerRecordBlockTransfer	= 0x37,		/* v1.2,		removed in 1.3a */
	CecTimerRecordBlockTransfer			= 0x38,		/* v1.2,		removed in 1.3a */
	CecEndTimerRecordBlockTransfer		= 0x39,		/* v1.2,		removed in 1.3a */
	CecGiveTimerOptions					= 0x3e,		/* v1.2,		removed in 1.3a */
	CecTimerOptions						= 0x3f,		/* v1.2,		removed in 1.3a */
	CecPlay								= 0x41,
	CecDeckControl						= 0x42,
	CecTimerClearedStatus				= 0x43,
	CecUserControlPressed				= 0x44,
	CecUserControlReleased				= 0x45,
	CecGiveOsdName						= 0x46,
	CecSetOsdName						= 0x47,
	CecGiveSystemInfo					= 0x52,		/* v1.2,		removed in 1.3a */
	CecSetSystemInfoVersionNumber		= 0x53,		/* v1.2,		removed in 1.3a */
	CecSetOsdString						= 0x64,
	CecSetTimerProgramTitle				= 0x67,
	CecSystemAudioModeRequest			= 0x70,		/* 1.3a */
	CecGiveAudioStatus					= 0x71,		/* 1.3a */
	CecSetSystemAudioMode				= 0x72,		/* 1.3a */
	CecReportAudioStatus				= 0x7a,		/* 1.3a */
	CecGiveSystemAudioModeStatus		= 0x7d,		/* 1.3a */
	CecSystemAudioModeStatus			= 0x7e,		/* 1.3a */
	CecRoutingChange					= 0x80,
	CecRoutingInformation				= 0x81,
	CecActiveSource						= 0x82,
	CecGivePhysicalAddress				= 0x83,
	CecReportPhysicalAddress			= 0x84,
	CecRequestActiveSource				= 0x85,
	CecSetStreamPath					= 0x86,
	CecDeviceVendorId					= 0x87,
	CecVendorCommand					= 0x89,
	CecVendorRemoteButtonDown			= 0x8a,
	CecVendorRemoteButtonUp				= 0x8b,
	CecGiveDeviceVendorId				= 0x8c,
	CecMenuRequest						= 0x8d,
	CecMenuStatus						= 0x8e,
	CecGiveDevicePowerStatus			= 0x8f,
	CecGetDevPowerStatus				= 0x8f,		/* Kept for backward compatibility */
	CecReportPowerStatus				= 0x90,
	CecGetMenuLanguage					= 0x91,		/* 1.3a */
	CecSelectAnalogueService			= 0x92,		/* 1.3a */
	CecSelectDigitalService				= 0x93,		/* 1.3a */
	CecSetDigitalTimer					= 0x97,		/* 1.3a */
	CecClearDigitalTimer				= 0x99,		/* 1.3a */
	CecSetAudioRate						= 0x9a,		/* 1.3a */
	CecInactiveSource					= 0x9d,		/* 1.3a */
	CecCecVersion						= 0x9e,		/* 1.3a */
	CecGetCecVersion					= 0x9f,		/* 1.3a */
	CecVendorCommandWithID				= 0xa0,		/* 1.3a */
	CecClearExternalTimer				= 0xa1,		/* 1.3a */
	CecSetExternalTimer					= 0xa2,		/* 1.3a */
	CecReportShortAudioDescriptor		= 0xa3,		/* 1.4a */
	CecRequestShortAudioDescriptor		= 0xa4,		/* 1.4a */
	CecPollMessage						= 0xaa,
	CecInitiateArc						= 0xc0,		/* 1.4a */
	CecReportArcInitiated				= 0xc1,		/* 1.4a */
	CecReportArcTerminated				= 0xc2,		/* 1.4a */
	CecRequestArcInitiation				= 0xc3,		/* 1.4a */
	CecRequestArcTermination			= 0xc4,		/* 1.4a */
	CecTerminateArc						= 0xc5,		/* 1.4a */
	CecGiveFeatures						= 0xa5,		/* 2.0 */
	CecReportFeatures					= 0xa6,		/* 2.0 */
	CecRequestCurrentLatency			= 0xa7,		/* 2.0 */
	CecReportCurrentLatency				= 0xa8,		/* 2.0 */
	CecCdcMessage						= 0xf8,		/* 1.4a */
	CecAbort							= 0xff,
	CecAbortMessage						= 0xff,		/* Kept for backward compatibility */
};


static Opcode Opcodes[] = {
	{ CecFeatureAbort,					2,	"feature abort" },
	{ CecSetAnaloguePreset,				1,	"set analogue preset" },
	{ CecImageViewOn,					0,	"image view on" },
	{ CecTunerStepIncrement,			0,	"tuner step increment" },
	{ CecTunerStepDecrement,			0,	"tuner step decrement" },
	{ CecTunerDeviceStatus,				1,	"tuner device status" },
	{ CecGiveTunerDeviceStatus,			-1,	"give tuner device status" },
	{ CecRecordOn,						1,	"record on" },
	{ CecRecordStatus,					1,	"record status" },
	{ CecRecordOff,						0,	"record off" },
	{ CecTextViewOn,					0,	"text view on" },
	{ CecRecordTvScreen,				0,	"record tv screen" },
	{ CecGiveDeckStatus,				1,	"give deck status" },
	{ CecDeckStatus,					1,	"deck status" },
	{ CecSetMenuLanguage,				3,	"set menu language" },
	{ CecClearAnalogueTimer,			11,	"clear analog timer" },
	{ CecSetAnalogueTimer,				11,	"set analog timer" },
	{ CecTimerStatus,					-1,	"timer status" },					/* 1 or 3 bytes */
	{ CecStandby,						0,	"standby" },
	{ CecPlay,							1,	"play" },
	{ CecDeckControl,					1,	"deck control" },
	{ CecTimerClearedStatus,			1,	"timer cleared status" },
	{ CecUserControlPressed,			1,	"user control pressed" },
	{ CecUserControlReleased,			0,	"user control released" },
	{ CecGiveOsdName,					0,	"give osd name" },
	{ CecSetOsdName,					-1,	"set osd name" },
	{ CecSetOsdString,					-1,	"set osd string" },
	{ CecSetTimerProgramTitle,			-1,	"set timer program title" },
	{ CecSystemAudioModeRequest,		-1,	"system audio mode request" },		/* length maybe 0 or 2 */
	{ CecGiveAudioStatus,				0,	"give audio status" },
	{ CecSetSystemAudioMode,			1,	"set system audio mode" },
	{ CecReportAudioStatus,				1,	"report audio status" },
	{ CecGiveSystemAudioModeStatus,		0,	"give system audio mode status" },
	{ CecSystemAudioModeStatus,			1,	"system audio mode status" },
	{ CecRoutingChange,					4,	"routing change" },
	{ CecRoutingInformation,			2,	"routing information" },
	{ CecActiveSource,					2,	"active source" },
	{ CecGivePhysicalAddress,			0,	"give physical address" },
	{ CecReportPhysicalAddress,			3,	"report physical address" },
	{ CecRequestActiveSource,			0,	"request active source" },
	{ CecSetStreamPath,					2,	"set stream path" },
	{ CecDeviceVendorId,				3,	"device vendor id" },
	{ CecVendorCommand,					-1,	"vendor command" },
	{ CecVendorRemoteButtonDown,		-1,	"vendor remote button down" },
	{ CecVendorRemoteButtonUp,			-1,	"vendor remote button up" },
	{ CecGiveDeviceVendorId,			-1,	"give device vendor id" },
	{ CecMenuRequest,					1,	"menu request" },
	{ CecMenuStatus,					1,	"menu status" },
	{ CecGiveDevicePowerStatus,			0,	"give device power status" },
	{ CecReportPowerStatus,				1,	"report power status" },
	{ CecGetMenuLanguage,				0,	"get menu language" },
	{ CecSelectAnalogueService,			1,	"select analog service" },
	{ CecSelectDigitalService,			7,	"select digital service" },
	{ CecSetDigitalTimer,				14,	"set digital timer" },
	{ CecClearDigitalTimer,				14,	"clear digital timer" },
	{ CecSetAudioRate,					1,	"set audio rate" },
	{ CecInactiveSource,				2,	"inactive source" },
	{ CecCecVersion,					1,	"cec version" },
	{ CecGetCecVersion,					0,	"get cec version" },
	{ CecVendorCommandWithID,			-1,	"vendor command with id" },
	{ CecClearExternalTimer,			11,	"clear external timer" },
	{ CecSetExternalTimer,				11,	"set external timer" },
	{ CecReportShortAudioDescriptor,	-1,	"report short audio descriptor" },
	{ CecRequestShortAudioDescriptor,	-1,	"request short audio descriptor" },

	/* these are HDMI 1.4 specific */
	{ CecInitiateArc,					0,	"initiate arc" },
	{ CecReportArcInitiated,			0,	"report arc initiated" },
	{ CecReportArcTerminated,			0,	"report arc terminated" },
	{ CecRequestArcInitiation,			0,	"request arc initiation" },
	{ CecRequestArcTermination,			0,	"request arc termination" },
	{ CecTerminateArc,					0,	"terminate arc" },
	{ CecCdcMessage,					0,	"cdc message" },

	/* these are only in HDMI 2.0 specification */
	{ CecGiveFeatures,					0,	"give current features" },
	{ CecReportFeatures,				0,	"report current features" },
	{ CecRequestCurrentLatency,			4,	"request current latency" },
	{ CecReportCurrentLatency,			0,	"report current latency" },

	{ CecAbort,							0,	"abort" }
};

static Map Keys[] = {
	{ 0x00,		"select" },
	{ 0x01,		"up" },
	{ 0x02,		"down" },
	{ 0x03,		"left" },
	{ 0x04,		"right" },
	{ 0x05,		"right up" },
	{ 0x06,		"right down" },
	{ 0x07,		"left up" },
	{ 0x08,		"left down" },
	{ 0x09,		"root menu" },
	{ 0x0a,		"setup menu" },
	{ 0x0b,		"contents menu" },
	{ 0x0c,		"favorite menu" },
	{ 0x0d,		"exit" },

	/* 0x0e-0x0f: Reserved */
	{ 0x10,		"media top level" },
	{ 0x11,		"media, content sensitive menu" },

	/* 0x12-0x1c: Reserved */
	{ 0x1d,		"number entry mode" },
	{ 0x1e,		"11" },
	{ 0x1f,		"12" },
	{ 0x20,		"0 or 10" },
	{ 0x21,		"1" },
	{ 0x22,		"2" },
	{ 0x23,		"3" },
	{ 0x24,		"4" },
	{ 0x25,		"5" },
	{ 0x26,		"6" },
	{ 0x27,		"7" },
	{ 0x28,		"8" },
	{ 0x29,		"9" },
	{ 0x2a,		"dot" },
	{ 0x2b,		"enter" },
	{ 0x2c,		"clear" },

	/* 0x2d-0x2e: Reserved */
	{ 0x2f,		"next favorite" },
	{ 0x30,		"channel up" },
	{ 0x31,		"channel down" },
	{ 0x32,		"previous channel" },
	{ 0x33,		"sound select" },
	{ 0x34,		"input select" },
	{ 0x35,		"display info" },
	{ 0x36,		"help" },
	{ 0x37,		"page up" },
	{ 0x38,		"page down" },

	/* 0x39-0x3f: Reserved */
	{ 0x40,		"power" },
	{ 0x41,		"volume up" },
	{ 0x42,		"volume down" },
	{ 0x43,		"mute" },
	{ 0x44,		"play" },
	{ 0x45,		"stop" },
	{ 0x46,		"pause" },
	{ 0x47,		"record" },
	{ 0x48,		"rewind" },
	{ 0x49,		"fast forward" },
	{ 0x4a,		"eject" },
	{ 0x4b,		"forward" },
	{ 0x4c,		"back" },
	{ 0x4d,		"stop record" },
	{ 0x4e,		"pause record" },

	/* 0x4f: Reserved */
	{ 0x50,		"angle" },
	{ 0x51,		"sub picture" },
	{ 0x52,		"video on demaqnd" },
	{ 0x53,		"electronic program guide" },
	{ 0x54,		"timer programming" },
	{ 0x55,		"initial config" },

	{ 0x56,		"select broadcast type" },		/* takes broadcast type, is non-deterministic */
	{ 0x57,		"select sound presentation" },	/* takes sound presentation control, is non-deterministic */

	/* 0x58-0x5f: Reserved */

	/*
	 * 0x60 - 0x6d are deterministic - the state after the press is
	 * well defined. Their names contain a '!' char to indicate this
	 */
	{ 0x60,		"play!" },					/* takes an aditional play mode param */
	{ 0x61,		"play pause!" },
	{ 0x62,		"record!" },
	{ 0x63,		"pause record!" },
	{ 0x64,		"stop!" },
	{ 0x65,		"mute!" },
	{ 0x66,		"un-mute!" },

	/* NB: these carry an aditional arg */
	{ 0x67,		"tune!" },					/* takes an aditional channel ID */
	{ 0x68,		"select media!" },			/* takes an aditional media ID */
	{ 0x69,		"select A/V input!" },		/* takes an aditional A/V input ID */
	{ 0x6a,		"select audio input!" },

	{ 0x6b,		"power toggle!" },
	{ 0x6c,		"power off!" },
	{ 0x6d,		"power on!" },

	/* 0x6e-0x70: Reserved */
	{ 0x71,		"f1/blue" },
	{ 0x72,		"f2/red" },
	{ 0x73,		"f3/green" },
	{ 0x74,		"f4/yellow" },
	{ 0x75,		"f5" },

	{ 0x76,		"data" },
	/* 0x77-0xff: Reserved */
};

static Array Devices[] = {
	"tv",
	"rec1",		/* recorder 1 */
	"rec2",		/* recorder 1 */
	"tun1",		/* tuner 1 */
	"pb1",		/* recorder 1 */
	"aud",		/* audio system */
	"tun2",		/* tuner 2 */
	"tun3",		/* tuner 3 */
	"pb2",		/* playback 2 */
	"rec3",		/* recorder 3 */
	"tun4",		/* tuner 4 */
	"pb3",		/* playback 3 */
	"res1",		/* reserved 1 */
	"res2",		/* reserved 2 */
	"spec",		/* specific use  */
	"unkn"		/* unregistered (as a source), broadcast (as a destination) */
};

static Map Vendors[] = {
	{ 0x000039,		"Toshiba 1" },
	{ 0x000CE7,		"Toshiba 2" },
	{ 0x0000F0,		"Samsung" },
	{ 0x0005CD,		"Denon" },
	{ 0x000678,		"Marantz" },
	{ 0x000982,		"Loewe" },
	{ 0x0009B0,		"Onkyo" },
	{ 0x000CB8,		"Medion" },
	{ 0x001582,		"Pulse eight" },
	{ 0x001950,		"Harman multimedia" },
	{ 0x9C645E,		"Harman consumer" },
	{ 0x001A11,		"Google" },
	{ 0x0020C7,		"Akai" },
	{ 0x002467,		"AOC" },
	{ 0x005060,		"Cisco" },
	{ 0x008045,		"Panasonic" },
	{ 0x00903E,		"Philips" },
	{ 0x009053,		"Daewoo" },
	{ 0x00A0DE,		"Yamaha" },
	{ 0x00D0D5,		"Grundig" },
	{ 0x00D38D,		"Hospitality profile" },		/* see "Hospitality Profile for HDMI CEC v1.3" */
	{ 0x00E036,		"Pioneer" },
	{ 0x00E091,		"LG" },
	{ 0x08001F,		"Sharp 1" },
	{ 0x534850,		"Sharp 2" },
	{ 0x080046,		"Sony" },
	{ 0x18C086,		"Broadcom" },
	{ 0x5cad76,		"TCL" },
	{ 0x6B746D,		"Vizio" },
	{ 0x743a65,		"NEC" },
	{ 0x8065E9,		"Benq" },
 	{ 0x001039,		"JVC" },

	/* unregistered (illegal) vendor IDs */
	{ 0x232425,		"Arcam" },
	{ 0x0005f7,		"CYP" },
	{ 0x00260f,		"Linn products" },
	{ 0x4e414c,		"Naim audio" },
};

static Map DeckStatus[] = {
	{ 0x11,	"play" },
	{ 0x12,	"record" },
	{ 0x13,	"play reverse" },
	{ 0x14,	"still" },
	{ 0x15,	"slow" },
	{ 0x16,	"slow reverse" },
	{ 0x17,	"fast forward" },
	{ 0x18,	"fast reverse" },
	{ 0x19,	"no media" },
	{ 0x1a,	"stop" },
	{ 0x1b,	"skip forward / wind" },
	{ 0x1c,	"skip reverse / rewind" },
	{ 0x1d,	"index search forward" },
	{ 0x1e,	"index search reverse" },
	{ 0x1f,	"other status" },
};

static Array Versions[] = {
	"v1.1",
	"v1.2",
	"v1.2a",
	"v1.3",
	"v1.3a",
	"v1.4a/b",
	"v2.0",
};

static Array GiveDeckStatus[] = {
	nil,
	"start",
	"stop",
	"once",
};


static Array Formats[] = {
	nil,
	"PCM",				/* linear PCM */
	"AC3",				/* AC3 */
	"MPEG-1 L1/L2",		/* Mpeg1 layer 1 or 2 */
	"MP3",				/* mpeg1 layer 3 */
	"MPEG2",			/* mpeg2 */
	"AAC",				/* AAC */
	"DTS",				/* DTS */
	"ATRAC",			/* Sony */
	"DSD",				/* Super audio CD, 1 bit */
	"eAC3",				/* Dolby Digital + */
	"DTS-HD",			/* DTS-HD */
	"MLP",				/* Meridian Lossless Packing */
	"DST",				/* Direct stream transfer */
	"WMA pro"			/* Microsoft WMA pro */
};

static Array Powerstates[] = {
	"active",
	"standby",
	"starting up",
	"shutting down"
};

static Array Aborts[] = {
	"unrecognised opcode",
	"not in the correct mode to respond",
	"cannot provide source",
	"invalid operand",
	"refused",
	"unable to determine"
};

static Array MenuState[] = {
	"active",
	"inactive",
};

static Array Boolean[] = {
	"false",
	"true",
};

static char *
lookoper(int id, int *nargs)
{
	int i;
	static char buf[24];

	for(i = 0; i < nelem(Opcodes); i++)
		if(id == Opcodes[i].id){
			if(nargs)
				*nargs = Opcodes[i].nargs;
			return Opcodes[i].str;
		}

	if(nargs)
		*nargs = -1;
	snprint(buf, sizeof(buf), "<unknown 0x%02x>", id);
	return buf;
}

char *
lookmap(Map *m, int n, int id)
{
	int i;
	static char buf[24];

	for(i = 0; i < n; i++, m++)
		if(m->id == id && m->str)
			return m->str;
	snprint(buf, sizeof(buf), "<unknown 0x%02x>", id);
	return buf;
}

char *
lookarray(Array *a, int n, int id)
{
	static char buf[24];

	if(id >= 0 && id < n && a[id])
		return a[id];
	snprint(buf, sizeof(buf), "<unknown 0x%02x>", id);
	return buf;
}

char *
device(int id, int is_dest)
{
	if(is_dest && id == 0x0f)		/* 0x0f is special, unregistered / broadcast */
		return "*";
	return lookarray(Devices, nelem(Devices), id);
}


void
dump(Biobuf *bp, uchar *buf, int len)
{
	char *name;
	int i, n, nargs;

	if(len == 0){
		Bprint(bp, "ping\n");
		return;
	}

	name = lookoper(buf[0], &nargs);
	if(len == 0)
		Bprint(bp, "%s", name);
	else
		Bprint(bp, "%-32s ", name);

	if(nargs != -1 && nargs != len-1){						/* damaged message */
		Bprint(bp, "<bad message length, %d != %d>\n", len-1, nargs);
		return;
	}

	/* Beware: This switch statement is incomplete, I decode only messages I have ever seen */

	switch(buf[0]){
	case CecFeatureAbort:
		Bprint(bp, "%q %q\n", lookoper(buf[1], nil), lookarray(Aborts, nelem(Aborts), buf[2]));
		break;
	case CecSetMenuLanguage:
		Bprint(bp, "'%c%c%c'\n", buf[1], buf[2], buf[3]);
		break;
	case CecUserControlPressed:
		Bprint(bp, "%q\n",  lookmap(Keys, nelem(Keys), buf[1]));
		break;
	case CecSetOsdName:
		Bprint(bp, "%.*q\n", len-1, (char *)buf+1);
		break;
	case CecSetOsdString:
		Bprint(bp, "%02x %.*q\n", buf[1], len-2, (char *)buf+2);
		break;
	case CecSetSystemAudioMode:
		Bprint(bp, "%s\n", lookarray(Boolean, nelem(Boolean), buf[1]));
		break;
	case CecSystemAudioModeStatus:
		Bprint(bp, "%s\n", lookarray(Boolean, nelem(Boolean), buf[1]));
		break;
	case CecReportAudioStatus:
		Bprint(bp, "mute=%s ",  lookarray(Boolean, nelem(Boolean), (buf[1] & 0x80)? 1: 0));
		if((buf[1] & 0x7f) == 127)
			Bprint(bp, " vol=?\n");
		if((buf[1] & 0x7f) > 100)
			Bprint(bp, " vol=0x%x <illegal value>\n", buf[1] & 0x7f);
		else
			Bprint(bp, " vol=%d%%\n", buf[1] & 0x7f);
		break;
	case CecSystemAudioModeRequest:
		switch(len-1){
		case 2:
			Bprint(bp, "@ %02x%02x <enable audio>\n", buf[1], buf[2]);
			break;
		case 0:
			Bprint(bp, "<disable audio>\n");
			break;
		default:
			Bprint(bp, "<%d bad arg count, expected 0 or 2>\n", len-1);
			break;
		}
		break;
	case CecRoutingChange:
		Bprint(bp, "@ %02x%02x → @ %02x%02x\n", buf[1], buf[2], buf[3], buf[4]);
		break;
	case CecSetStreamPath:
		Bprint(bp, "@ %02x%02x\n", buf[1], buf[2]);
		break;
	case CecGiveDeckStatus:
		Bprint(bp, "%q\n", lookarray(GiveDeckStatus, nelem(GiveDeckStatus), buf[1]));
		break;
	case CecDeckStatus:
		Bprint(bp, "%q\n", lookmap(DeckStatus, nelem(DeckStatus), buf[1]));
		break;
	case CecMenuStatus:
		Bprint(bp, "%q\n", lookarray(MenuState, nelem(MenuState), buf[1]));
		break;
	case CecActiveSource:
		Bprint(bp, "@ %02x%02x\n", buf[1], buf[2]);
		break;
	case CecInactiveSource:
		Bprint(bp, "@ %02x%02x\n", buf[1], buf[2]);
		break;
	case CecReportPhysicalAddress:
		Bprint(bp, "%s @ %02x%02x\n", lookarray(Devices, nelem(Devices), buf[3]), buf[1], buf[2]);
		break;
	case CecReportPowerStatus:
		Bprint(bp, "%q\n", lookarray(Powerstates, nelem(Powerstates), buf[1]));
		break;
	case CecCecVersion:
		Bprint(bp, "%q\n", lookarray(Versions, nelem(Versions), buf[1]));
		break;
	case CecDeviceVendorId:
		n = (buf[1]<<16)|(buf[2]<<8)|buf[3];
		Bprint(bp, "%q\n", lookmap(Vendors, nelem(Vendors), n));
		break;
	case CecReportShortAudioDescriptor:
		for(i = 1; i < len; i++)
			Bprint(bp, "%q ", lookarray(Formats, nelem(Formats), buf[i]));
		Bprint(bp, "\n");
		break;
	case CecRequestShortAudioDescriptor:
		for(i = 1; i < len; i++)
			Bprint(bp, "%q ", lookarray(Formats, nelem(Formats), buf[i]));
		Bprint(bp, "\n");
		break;
	case CecVendorCommandWithID:
		if(len > 4){
			n = (buf[1]<<16)|(buf[2]<<8)|buf[3];
			Bprint(bp, "%q [", lookmap(Vendors, nelem(Vendors), n));
			for(i = 4; i < len-1; i++)
				Bprint(bp, "%02x ", buf[i]);
			Bprint(bp, "%02x]\n", buf[i]);
		}
		else
		if(len > 1){
			Bprint(bp, "[");
			for(i = 1; i < len-1; i++)
				Bprint(bp, "%02x ", buf[i]);
			Bprint(bp, "%02x]\n", buf[i]);

		}
		break;
	default:
		if(len > 1){
			Bprint(bp, "[");
			for(i = 1; i < len-1; i++)
				Bprint(bp, "%02x ", buf[i]);
			Bprint(bp, "%02x]\n", buf[i]);
		}
		else
			Bprint(bp, "\n");
		break;
	}
}

void
usage(void)
{
	fprint(2, "usage: %s [-t] -l <n> [<file>]\n", argv0);
	fprint(2, "   -a        print all traffic, not just that which relates to us\n");
	fprint(2, "   -h        print source hex with decoded data\n");
	fprint(2, "   -l <n>    set the local logical address to <n>, default = 5\n");
	fprint(2, "   -t        first field on each line is a timestamp\n");
	fprint(2, "   -r        re-base time to zero at start of log\n");
	exits("usage");
}

void
main(int argc, char *argv[])
{
	long addrs;
	double first, when;
	char *line, *arr[MTU];
	uchar src, dst, buf[MTU];
	int len, all, loc, i, off, hex, rebase;
	Biobuf inp, *ip = &inp, out, *op = &out;

	Binit(op, 1, OWRITE);

	loc = 5;					/* assume we are an audio device */
	off = 0;
	hex = 0;
	all = 0;
	first = -1;
	rebase = 0;

	ARGBEGIN{
	case 'a':
		all++;
		break;
	case 'h':
		hex = 1;
		break;
	case 'l':
		loc = strtol(EARGF(usage()), nil, 0);
		break;
	case 'r':
		rebase = 1;
		break;
	case 't':
		off = 1;
		break;
	default:
		usage();
	}ARGEND;

	switch(argc){
	case 0:
		Binit(ip, 0, OREAD);
		break;
	case 1:
		if((ip = Bopen(argv[1], OREAD)) == nil)
			sysfatal("%s cannot open, %r", argv[1]);
		break;
	default:
		usage();
	}

	quotefmtinstall();

	while((line = Brdline(ip, '\n')) != nil){
		line[Blinelen(ip)-1] = 0;
		len = tokenize(line, arr, nelem(arr));

		if(len == 0 || arr[0][0] == '#'){
			for(i = 0; i < len; i++)
				Bprint(op, "%s ", arr[i]);
			Bprint(op, "\n");
		}

		if(len < off)
			continue;

		/* peek into the message to see if its source or dest is of interest */
		addrs = strtol(arr[off], nil, 16);
		src = (addrs>>4) & 0x0f;
		dst = addrs & 0x0f;
		if(!all && src != loc && dst != loc && dst != Broadcast)
			continue;

		/* extract timestamps from the start of incomming lines */
		if(off){
			when = atof(arr[0]);
			if(rebase){
				if(first == -1)
					first = when;
				when -= first;
			}
			Bprint(op, "%-10.3f    ", when);
		}

		/* print hexdumps of messages */
		if(hex){
			for(i = 0; i < len -off; i++)
				Bprint(op, "%02lx ",  strtol(arr[i+off], nil, 16));
			for(; i < MTU; i++)
				Bprint(op, "   ");
		}

		for(i = 0; i < len -off; i++)
			buf[i] = strtol(arr[i+off], nil, 16);

		if(src == loc)
			Bprint(op, "%4s ← %-4s    ", device(dst, 1), device(src, 0));
		else
			Bprint(op, "%4s → %-4s    ", device(src, 0), device(dst, 1));

		dump(op, buf +1, len -1 -off);
	}
	exits(nil);
}
